program jcInterpolation_v4

  ! Jason P. Chang
  ! July 2012
  ! 
  ! Based on Bob's interpolation code (IE. MUCH BETTER THAN YOUR OTHER VERSIONS)

  use sep

  implicit none
  
  ! allocate variables
  integer                                :: ii,ig1,ig2,i1,i2
  integer                                :: ix,iy,ia,ib
  integer                                :: xindx,yindx
  integer                                :: nmax,nmaxdist,ilevel,nf
  integer                                :: n1,n2
  integer                                :: n1g,n2g
  integer                                :: n1xy,n2xy
  integer,allocatable,dimension (:,:)    :: count,count2
  integer,allocatable,dimension (:,:,:)  :: map,map2
  real                                   :: scale,cx,cy
  real                                   :: o1g,o2g,d1g,d2g,d1,o1
  real   ,allocatable,dimension (:,:)    :: header,dd
  real   ,allocatable,dimension (:,:,:)  :: mm,scale2
  logical                                :: verb,keep,normalize

  ! open I/O
  call sep_init()

  call from_param('verb',verb,.false.)
  call from_param('normalize',normalize,.true.)

  ! call in stuff from actual history file
  call from_history('n1',n1,1 )
  call from_history('n2',n2,1 )
  call from_history('d1',d1,1.)
  call from_history('o1',o1,1.)

  ! call in stuff from gridding parameter file
  call from_param('ng1',n1g,1 )
  call from_param('ng2',n2g,1 )
  call from_param('og1',o1g,1.)
  call from_param('og2',o2g,1.)
  call from_param('dg1',d1g,1.)
  call from_param('dg2',d2g,1.)

  ! call in stuff from header file (for station locations)
  call from_param('xindx',xindx,1)
  call from_param('yindx',yindx,2)
  call from_aux('xy_in','n1',n1xy)
  call from_aux('xy_in','n2',n2xy)
  if ( n2 .ne. n2xy ) call erexit ("n2 and n2xy do not match")

  ! call out
  call to_history('n1',n1g)
  call to_history('n2',n2g)
  call to_history('n3',n1 )
  call to_history('d1',d1g)
  call to_history('d2',d2g)
  call to_history('d3',d1 )
  call to_history('o1',o1g)
  call to_history('o2',o2g)
  call to_history('o3',o1 )

  call sep_close()

  ! allocation of model, data, header information
  allocate( mm    (n1g,n2g,n1) ) ; mm=0.
  allocate( dd    (n1  ,n2   ) )
  allocate( header(n1xy,n2xy ) )

  ! allocate map and fold
  allocate( map  (5,n1g,n2g) ) ; map = -1
  allocate( count(  n1g,n2g) ) ; count = 0

  ! read in the station data and header information
  ! do this outside should adjoint have to work later
  do ii=1,n2
    call sep_read(dd(:,ii))
  end do
  call sep_read(header,'xy_in')

  ! NORMALIZE THE AMPLITUDES (SINCE SOME ARE ANOMALOUSLY HIGH)
  if (normalize) then
  ! loop over stations
  do ii=1,n2
    scale = sum( abs(dd(:,ii)) )

    if ( scale .gt. .00001 ) then
      dd(:,ii) = dd(:,ii) / scale
      if ( verb ) write(0,*) "scaling trace",ii,"by",1./scale
    end if
  end do
  end if

  ! MAP POINTS TO CLOSEST BIN (max 5 per bin)
  do ii=1,n2
    ix = 1.5 + ( header(xindx,ii) - o1g ) / d1g
    iy = 1.5 + ( header(yindx,ii) - o2g ) / d2g
    if ( (1 .gt. ix) .or. (ix .gt. n1g) .or. (1 .gt. iy) .or. (iy .gt. n2g) ) call erexit("Grid dimensions do not work")
    
    if ( count(ix,iy) .lt. 5 ) then
      count(ix,iy) = count(ix,iy) + 1
      map(count(ix,iy),ix,iy) = ii
    end if
  end do

  ! interpolation maximum bin distance
  nmax = 0
  nmaxdist = 4

  ! LOOP OVER THE BINS, FIND THE MAXIMUM NUMBER OF TRACES NEED FOR INTERPOLATION (NO DATA HANDLING YET STILL)
  ! description:
  !   - sums over a square whose size is determined by ilevel
  !   - ilevel    : size of bin search (eg. level 0 is within the same bin, level 1 is one bin over in all directions)
  !   - nmaxdist  : number of bins to search across in a given direction
  !   - nf        : the number of stations to be used for each grid cell (max and min stuff are boundary checks)
  !   - nmax      : maximum number of traces that will ever be needed for interpolating (THE GOAL OF THIS SECTION)
  do ig1=1,n1g
    do ig2=1,n2g
      keep = .false.
      ilevel = 0
      !loop up to 3 bins to either side looking for close by
      do while( ilevel < nmaxdist .and. .not. keep ) 
        nf = sum(count(max(1,ig1-ilevel):min(n1g,ig1+ilevel),max(1,ig2-ilevel):min(n2g,ig2+ilevel)))
        if( nf .gt. 0 ) then
          nmax = max(nmax,nf)
          keep = .true.
        else
          ilevel = ilevel + 1
        end if
      end do
    end do
  end do
  if ( verb ) write(0,*) "MAXIMUM COUNT",nmax,n1g,n2g

  ! NOW ACTUALLY COMBINE THE NEARBY TRACES (GET STATION INDICES)
  ! recall:
  !   - old map:   station indices (else -1)
  !   - old count: number of stations in each grid cell (else 0)
  ! count2: number of traces to grab for each cell

  ! allocate
  allocate( count2(n1g,n2g     ) ) ; count2 = 0
  allocate( map2  (nmax,n1g,n2g) ) ; map2 = 0
  allocate( scale2(nmax,n1g,n2g) ) ; scale2 = 0

  ! loop over bins again
  do ig1=1,n1g
    do ig2=1,n2g
      keep = .false.
      ilevel = 0
      !loop up to 3 bins to either side looking for close by
      do while( ilevel < nmaxdist .and. .not. keep ) 
        nf = sum(count(max(1,ig1-ilevel):min(n1g,ig1+ilevel),max(1,ig2-ilevel):min(n2g,ig2+ilevel)))
        if (nf .gt. 0) then
          ! loop over y-range search box
          do ib=max(1,ig2-ilevel),min(n2g,ig2+ilevel)
            ! loop over x-range search box
            do ia=max(1,ig1-ilevel),min(n1g,ig1+ilevel)
              ! smart way of summing all station indices from each cell into an interpolation grid (COULD BE IG1 RATHER THAN IX)
              map2(count2(ig1,ig2)+1:count2(ig1,ig2)+count(ia,ib),ig1,ig2) = map(:,ia,ib)
              count2(ig1,ig2) = count2(ig1,ig2) + count(ia,ib)
            end do
          end do
          keep = .true.
        else
          ilevel = ilevel + 1
        end if
      end do
    end do
  end do
  if ( verb ) write(0,*) "FINISHED COLLECTING NOW TIME TO INTERPOLATE"

  !NOW it is time to find the weights
  do ig2=1,n2g
    cy = (ig2-1)*d2g + o2g
    do ig1=1,n1g
      cx = (ig1-1)*d1g + o1g

      ! to not fill in empty bins
      if ( count(ig1,ig2) .eq. 0. ) then
        !do nothing
      else

        ! didn't find any nearby points
        if( count2(ig1,ig2) .eq. 0) then
          !do nothing
        else
          ! we will just do averaging based on inverse imaging
          if ( verb ) write(0,*) "here we are",ig1,ig2,count2(ig1,ig2)
          do i1=1,count2(ig1,ig2)
            ! note the d2g/20. factor to correct for points that fall exactly on the grid cell
            scale2(i1,ig1,ig2) = 1./(d2g/20.+&
              sqrt((cx-header(xindx,map2(i1,ig1,ig2)))**2+&
               (cy-header(yindx,map2(i1,ig1,ig2)))**2))
          end do

          if ( verb ) write(0,*) "OUT OF DO",count2(ig1,ig2),sum(scale2(:count2(ig1,ig2),ig1,ig2))
          !scale by the binning weights
          scale2(:count2(ig1,ig2),ig1,ig2) = scale2(:count2(ig1,ig2),ig1,ig2) / sum(scale2(:count2(ig1,ig2),ig1,ig2))
          if ( verb ) write(0,*) "found",scale2(:5,ig1,ig2)
          do i2=1,count2(ig1,ig2)
            mm(ig1,ig2,:) = mm(ig1,ig2,:) + scale2(i2,ig1,ig2)*dd(:,map2(i2,ig1,ig2))
          end do
        end if
      end if
    end do
  end do

  ! need to write out here to save memory
  call sep_write(mm)

end program jcInterpolation_v4
